#!/usr/bin/env python3

# standards
from doctest import DocTestParser, DocTestRunner
from pathlib import Path
import re

# poisk
import poisk


def test_readme_examples():
    readme_file = Path(__file__).parent / '..' / 'README.md'
    doctest_str = '\n\n'.join(
        re.sub(
            # We auto-insert doctest <BLANKLINE> markers into the examples. Don't want to put them in the README itself, it would
            # make the examples a bit confusing.
            r'\n\n(?!>>>)',
            '\n<BLANKLINE>\n',
            block_str
        )
        for block_str in re.findall(
            r'```python\s+(.+?)```',
            readme_file.read_text('UTF-8'),
            flags=re.S,
        )
    )
    assert doctest_str
    print(doctest_str)
    parser = DocTestParser()
    runner = DocTestRunner()
    runner.run(
        parser.get_doctest(
            doctest_str,
            dict(globals(), poisk=poisk, __file__=readme_file),
            'README.md',
            'README.md',
            0,
        ),
    )
    assert runner.failures == 0



#!/usr/bin/env python3

# standards
from pathlib import Path
from textwrap import dedent

# 3rd parties
from mypy.build import build  # pylint: disable=no-name-in-module
from mypy.modulefinder import BuildSource  # pylint: disable=no-name-in-module
from mypy.options import Options  # pylint: disable=no-name-in-module
import pytest
import yaml


def load_fixtures():
    auto_imports = dedent('''\
        import re
        import lxml.etree
        from poisk import many, one
    ''')
    fixtures_file = Path(__file__).parent / 'test_type_checking.yaml'
    doc = yaml.safe_load(fixtures_file.read_text('UTF-8'))
    for test_case in doc['test_cases']:
        yield (
            test_case['name'],
            auto_imports + test_case['code'],
            test_case['expected_error'],
        )


@pytest.mark.parametrize(
    'code_str, expected_error',
    [
        pytest.param(code_str, expected_error, id=name)
        for name, code_str, expected_error in load_fixtures()
    ],
)
def test_type_checking(code_str, expected_error):
    """ Check that mypy can recognise correct invocations of poisk. """
    options = Options()
    options.show_error_codes = True
    options.ignore_missing_imports = True
    all_errors = []
    build(
        [BuildSource(None, None, code_str)],
        options,
        flush_errors=lambda errors, *rest, **kwargs: all_errors.extend(errors),
    )
    if expected_error is None:
        if all_errors:
            pytest.fail('\n'.join(all_errors))
    else:
        assert expected_error in '\n'.join(all_errors)



#!/usr/bin/env python3

"""
Implements a function for searching in Plain Old Data Structures.


Initially this was going to use JMESPath, but that library has a few limitations that made it tricky to use here:

<> We can't easily tell these two apart:

       >>> jmespath.search("list", {"list": [1, 2]})
       [1, 2]
       >>> jmespath.search("list[]", {"list": [1, 2]})
       [1, 2]

   But the difference matters to us here, because a `one.pods` of the list should work (you're finding the one list at that path),
   whereas a `one.pods` of the second should raise `ManyFound` (since the list has more than one element).

<> We can't tell these apart either:

       >>> jmespath.search("myvalue", {"myvalue": None})
       None
       >>> jmespath.search("myvalue", {})
       None

   But again the difference matters to us. `one.pods` should return `None` in the first case, but raise `NotFound` in the second.

So we hack our own, and it does the job.


Note that this is therefore a lot simpler and much less featured than JMESPath. For instance we don't do filters like
"locations[?state == 'WA'].name", we don't have the pipe operator, functions like `sort` etc. But these aren't really necessary
when working in Python, as you can just as easily write those as list expressions or function calls.
"""


# standards
from collections.abc import Mapping, Sequence
import re
from typing import Iterable, List, Mapping as MappingType, Type, TypeVar, Union, overload


CHILDREN = object()


T = TypeVar('T')  # pylint: disable=invalid-name

SearchablePods = Union[MappingType, list, tuple]  # NB not using `Sequence` as we don't want to include `str`


@overload
def pods_search(
    needle: str,
    haystack: SearchablePods,
) -> List[object]:
    ...

@overload
def pods_search(
    needle: str,
    haystack: SearchablePods,
    type: Type[T],
) -> List[T]:
    ...


def pods_search(
    needle: str,
    haystack: SearchablePods,
    type = None,
):
    results = []
    stack = [(haystack, list(_parse_steps(needle)))]
    while stack:
        node, steps = stack.pop()
        if not steps:
            if type is not None and not isinstance(node, type):
                raise TypeError(f'Expected {type.__name__}, found {node.__class__.__name__}')
            results.append(node)
        else:
            head, *tail = steps
            if head is CHILDREN:
                if isinstance(node, Sequence) and not isinstance(node, str):
                    for element in reversed(node):
                        stack.append((element, tail))
            elif (isinstance(node, Mapping) and head in node) or (isinstance(node, Sequence) and isinstance(head, int)):
                stack.append((node[head], tail))  # type: ignore  # mypy gets confused but I think it's fine
    return results


def _parse_steps(needle: str) -> Iterable[Union[object]]:
    """
    Splits the user-specified `needle` into its components. See examples of the supported format in `test_poisk.py`.

    >>> list(_parse_steps("a.b.c"))
    ["a", "b", "c"]
    >>> list(_parse_steps("a[].b[0].c"))
    ["a", CHILDREN, "b", 0, "c"]
    """
    pos = 0
    re_step = re.compile(
        r'''
          \s*
          (?:
              "  (?P<double> (?:[^\\"]|\\.)+ ) "
            | '  (?P<single> (?:[^\\']|\\.)+ ) '
            |    (?P<word> [\w\-\$]+ )
            | \[ (?P<index> \d+ ) \]
            |    (?P<brackets> \[\] )
          )
          (?: \s*\.\s* | (?=\s*\[) | $ )
        ''',
        flags=re.X,
    )
    while pos < len(needle):
        match = re_step.match(needle, pos)
        if not match:
            raise ValueError(f"Can't parse needle at '{needle[pos:]}'")
        groups = match.groupdict()
        if groups.get('double') or groups.get('single'):
            string: str = groups.get('double') or groups.get('single')  # type: ignore[assignment]
            yield re.sub(r'\\(.)', r'\1', string)
        elif groups.get('index'):
            yield int(groups['index'])
        elif groups.get('brackets'):
            yield CHILDREN
        else:
            yield groups.get('word')
        pos = match.end()



#!/usr/bin/env python3

# pylint: disable=line-too-long

# standards
from collections.abc import Mapping, Sequence
import re

# 3rd parties
import lxml.etree as ET
import pytest

# poisk
from poisk import ManyFound, NotFound, many, one


HTML_DOC = ET.HTML(
    '''
    <html>
      <body>
        <p id="first">Au large, <b>forban</b>!</p>
        <p class="second">Au large, flibustier!</p>
      </body>
    </html>
    '''
)


FIRST_P = one.etree(
    '#one',
    ET.fromstring(
        '''
        <div>
          <p id="one">This is <b>one has <b>nests</b></b></p>
          <p id="two">This is <b>two</b></p>
        ''',
        parser=ET.HTMLParser(),
    )
)


XML_DOC = ET.XML(
    '''
    <doc
      xmlns:a="http://xml.com/ns1"
      xmlns:b="http://xml.com/ns2"
    >
      <a:foo>
         <b:bar>Text</b:bar>
      </a:foo>
    </doc>
    '''
)


class MyClass:
    # pylint: disable=too-few-public-methods
    pass


class MyString(str):
    # pylint: disable=too-few-public-methods
    pass


class MySequence(Sequence):

    def __init__(self, *values):
        self.values = values

    def __getitem__(self, item):
        return self.values[item]

    def __len__(self):
        return len(self.values)


class MyMapping(Mapping):

    def __init__(self, **items):
        self.items = items

    def __getitem__(self, item):
        return self.items[item]

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)


class MyTree:
    # pylint: disable=too-few-public-methods

    def __init__(self, element):
        self.element = element

    def xpath(self, *args, **kwargs):
        return self.element.xpath(*args, **kwargs)


@pytest.mark.parametrize(
    'find, haystack, needle, options, expected',
    [

        # regex matching
        (one.re, 'abracadabra', r'br.c', {}, 'brac'),
        (one.re, 'abracadabra', r'br(.)c', {}, 'a'),
        (one.re, 'abracadabra', r'a(.+)a(.+)a(.+)a(.+)a', {}, ('br', 'c', 'd', 'br')),
        (one.re, 'abracadabra', r'b.a', {}, ManyFound),
        (one.re, 'abracadabra', r'b.a', {'allow_many': True}, 'bra'),
        (one.re, 'abracadabra', r'brr', {}, NotFound),
        (one.re, 'abracadabra', r'brr', {'allow_mismatch': True}, None),
        (one.re, 'abracadabra', r'BR.C', {}, NotFound),

        # regex matching kwags
        (one.re, 'abracadabra', r'B R . C', {}, NotFound),
        (one.re, 'abracadabra', r'B R . C', {'flags': re.I|re.X}, 'brac'),
        (one.re, 'abracadabra', 'a', {'made_up_kwarg': True}, TypeError),

        # subclassing a string works
        (one.re, MyString('hello'), r'.+(?=.)', {}, 'hell'),

        # `parse` is applied iff a match was found
        (one.re, 'I have 10 brothers', r'\d+', {'parse': int}, 10),
        (one.re, 'I have no brothers', r'\d+', {'parse': int, 'allow_mismatch': True}, None),

        # pods matching
        (one.pods, {'number': 1926}, 'number', {}, 1926),
        (one.pods, {'string': 's'}, 'string', {}, 's'),
        (one.pods, {'dict': {'k': 'v'}}, 'dict', {}, {'k': 'v'}),
        (one.pods, {'null': None}, 'null', {}, None),
        (one.pods, {}, 'anything', {}, NotFound),
        (one.pods, {}, 'anything', {'allow_mismatch': True}, None),
        (one.pods, [{'number': 1926}], '[].number', {}, 1926),
        # if the needle resolves to a single list, we return the list. Empty lists are fine
        (one.pods, {'list': []}, 'list', {}, []),
        (one.pods, {'list': [1, 2, 3]}, 'list', {}, [1, 2, 3]),
        # if the needle ends in `[]`, however, meaning that we intent to select the list's elements, then the list must have a
        # single element
        (one.pods, {'list': []}, 'list[]', {}, NotFound),
        (one.pods, {'list': []}, 'list[]', {'allow_mismatch': True}, None),
        (one.pods, {'list': [1]}, 'list[]', {}, 1),
        (one.pods, {'list': [1, 2]}, 'list[]', {}, ManyFound),
        (one.pods, {'list': [1, 2]}, 'list[]', {'allow_many': True}, 1),
        (one.pods, {'list': [[1, 2]]}, 'list[]', {}, [1, 2]),
        (one.pods, {'list_of_obj': [{'v': 1}, {'v': 2}]}, 'list_of_obj', {}, [{'v': 1}, {'v': 2}]),
        (one.pods, {'list_of_obj': [{'v': 1}, {'v': 2}]}, 'list_of_obj[]', {}, ManyFound),
        (one.pods, {'list_of_obj': [{'v': 1}, {'v': 2}]}, 'list_of_obj[].v', {}, ManyFound),
        (one.pods, {'list_of_obj': [{'v': 1}, {'other': 2}]}, 'list_of_obj[].v', {}, 1),
        (one.pods, {'list_of_obj': [{'v': 1}, {'other': 2}]}, 'list_of_obj[].k', {}, NotFound),
        (one.pods, {'one two': 12}, 'one two', {}, ValueError),  # can't parse the needle
        (one.pods, {'one two': 12}, '"one two"', {}, 12),
        (one.pods, {'one two': 12}, "'one two'", {}, 12),
        (one.pods, {'"hello"': 12}, '\'"hello"\'', {}, 12),
        # any sequence can be traversed the same as a list
        (one.pods, {'tuple': ({'v': 1},)}, 'tuple[].v', {}, 1),
        (one.pods, {'seq': MySequence({'v': 1})}, 'seq[].v', {}, 1),
        (one.pods, MyMapping(x=MySequence(12)), 'x[]', {}, 12),
        # '[]' needs to be quoted
        (one.pods, {'list[]': [1, 2, 3]}, 'list[]', {}, NotFound),
        (one.pods, {'list[]': [1, 2, 3]}, '"list[]"', {}, [1, 2, 3]),
        # list index
        (one.pods, {'n': [0, 1, 2]}, 'n[0]', {}, 0),
        # type checking
        (one.pods, {'list': [1, 2, 3]}, 'list', {'type': list}, [1, 2, 3]),
        (one.pods, {'list': [1, 2, 3]}, 'list', {'type': int}, TypeError),
        (one.pods, {'list': []}, 'list', {'type': list}, []),
        (one.pods, {'list': []}, 'list[]', {'type': list}, NotFound),

        # xpath matching
        (one.etree, HTML_DOC, 'body/p/b/text()', {}, 'forban'),
        (one.etree, HTML_DOC, './body/p/b/text()', {}, 'forban'),
        (one.etree, HTML_DOC, '/body/p/b/text()', {}, 'forban'),
        (one.etree, HTML_DOC, '//body/p/b/text()', {}, 'forban'),
        (one.etree, HTML_DOC, 'b/text()', {}, 'forban'),
        (one.etree, HTML_DOC, '//b/text()', {}, 'forban'),
        (one.etree, HTML_DOC, 'i/text()', {}, NotFound),
        (one.etree, HTML_DOC, 'i/text()', {'allow_mismatch': True}, None),
        (one.etree, HTML_DOC, 'p/text()', {}, ManyFound),
        (one.etree, HTML_DOC, 'p/text()', {'allow_many': True}, 'Au large, '),

        # Element.xpath() kwargs, used for namespaces, or passing arbitrary variables, see https://lxml.de/xpathxslt.html
        (one.etree, HTML_DOC, '//p[@id = $my_var]/b', {'my_var': 'first'}, '<b>forban</b>!'),
        (one.etree, XML_DOC, 'foo/bar/text()', {}, NotFound),
        (one.etree, XML_DOC, 'x:foo/y:bar/text()', {'namespaces': {'x': 'http://wrong.com/blah', 'y': 'http://other.com/boo'}}, NotFound),
        (one.etree, XML_DOC, 'x:foo/y:bar/text()', {'namespaces': {'x': 'http://xml.com/ns1', 'y': 'http://xml.com/ns2'}}, 'Text'),

        # xpath matching is limited to the given node's subtree
        (one.etree, FIRST_P.getparent(), 'b', {}, ManyFound),
        (one.etree, FIRST_P, 'b', {}, ManyFound),
        (one.etree, FIRST_P, '/b', {}, '<b>one has <b>nests</b></b>'),
        (one.etree, FIRST_P, './b', {}, '<b>one has <b>nests</b></b>'),
        (one.etree, FIRST_P, '//b', {}, ManyFound),
        (one.etree, FIRST_P, './/b', {}, ManyFound),

        # css matching
        (one.etree, HTML_DOC, 'p b', {}, '<b>forban</b>!'),
        (one.etree, HTML_DOC, 'p > b', {}, '<b>forban</b>!'),
        (one.etree, HTML_DOC, 'body b', {}, '<b>forban</b>!'),
        (one.etree, HTML_DOC, 'body > b', {}, NotFound),
        (one.etree, HTML_DOC, 'body > b', {'allow_mismatch': True}, None),
        (one.etree, HTML_DOC, 'p', {}, ManyFound),
        (one.etree, HTML_DOC, 'p', {'allow_many': True}, '<p id="first">Au large, <b>forban</b>!</p>'),
        (one.etree, HTML_DOC, 'p#first b', {}, '<b>forban</b>!'),
        (one.etree, HTML_DOC, 'p.second', {}, '<p class="second">Au large, flibustier!</p>'),

        # accessing attributes using xpath
        (one.etree, FIRST_P, '@id', {}, 'one'),
        (one.etree, FIRST_P, './@id', {}, 'one'),
        (one.etree, FIRST_P, '/@id', {}, 'one'),

        # anything with an 'xpath' method works
        (one.etree, MyTree(HTML_DOC), 'body/p/b/text()', {}, 'forban'),
        (one.etree, MyTree(HTML_DOC), 'p b', {}, '<b>forban</b>!'),

        # arbitrary list filtering
        (one.filter, ['', None, False, 'boo'], bool, {}, 'boo'),
        (one.filter, ['', None, False, 'boo'], lambda v: v == 'boom', {}, NotFound),
        (one.filter, ['', None, False, 'boo'], lambda v: v == 'boom', {'allow_mismatch': True}, None),
        (one.filter, ['', None, False, 'boo'], lambda v: not v, {}, ManyFound),
        (one.filter, ['', None, False, 'boo'], lambda v: not v, {'allow_many': True}, ''),

        # list filtering takes no kwargs
        (one.filter, [1, 2, 3], lambda v: v, {'anything_at_all': True}, TypeError),
    ]
)
def test_find_one(find, haystack, needle, options, expected):
    try:
        result = find(needle, haystack, **options)
    except Exception as ex:  # anything at all, pylint: disable=broad-except
        if isinstance(expected, type) and issubclass(expected, Exception):
            assert isinstance(ex, expected)
        else:
            raise
    else:
        if isinstance(result, ET._Element):  # pylint: disable=protected-access
            result = ET.tostring(result, encoding=str).strip()
        assert result == expected


@pytest.mark.parametrize(
    'find, haystack, needle, options, expected',
    [

        # regex matching
        (many.re, 'abracadabra', r'.a.', {}, ['rac', 'dab']),
        (many.re, 'abracadabra', r'(.a).', {}, ['ra', 'da']),
        (many.re, 'abracadabra', r'(.)a(.)', {}, [('r', 'c'), ('d', 'b')]),
        (many.re, 'abracadabra', r'z', {}, NotFound),
        (many.re, 'abracadabra', r'z', {'allow_mismatch': True}, []),

        # regex flags
        (many.re, 'abracadabra', r'. A .', {}, NotFound),
        (many.re, 'abracadabra', r'. A .', {'flags': re.I|re.X}, ['rac', 'dab']),
        (many.re, 'abracadabra', 'a', {'made_up_kwarg': True}, TypeError),

        # regex objects are also accepted
        (many.re, 'abracadabra', re.compile(r'.a.'), {}, ['rac', 'dab']),

        # subclassing a string works
        (many.re, MyString('hello'), r'[aeiou]', {}, ['e', 'o']),

        # `parse` is applied to every match
        (many.re, 'in my honest opinion', r'\b\w', {'parse': str.upper}, ['I', 'M', 'H', 'O']),

        # xpath matching
        (many.etree, HTML_DOC, 'body/p/text()', {}, ['Au large, ', '!', 'Au large, flibustier!']),
        (many.etree, HTML_DOC, 'body/div/text()', {}, NotFound),
        (many.etree, HTML_DOC, 'body/div/text()', {'allow_mismatch': True}, []),

        # xpath matching is limited to the given node's subtree
        (many.etree, FIRST_P.getparent(), 'b', {}, ['<b>one has <b>nests</b></b>', '<b>nests</b>', '<b>two</b>']),
        (many.etree, FIRST_P, 'b', {}, ['<b>one has <b>nests</b></b>', '<b>nests</b>']),
        (many.etree, FIRST_P, '/b', {}, ['<b>one has <b>nests</b></b>']),
        (many.etree, FIRST_P, './b', {}, ['<b>one has <b>nests</b></b>']),
        (many.etree, FIRST_P, '//b', {}, ['<b>one has <b>nests</b></b>', '<b>nests</b>']),
        (many.etree, FIRST_P, './/b', {}, ['<b>one has <b>nests</b></b>', '<b>nests</b>']),

        # css matching
        (many.etree, HTML_DOC, 'p', {}, ['<p id="first">Au large, <b>forban</b>!</p>', '<p class="second">Au large, flibustier!</p>']),
        (many.etree, HTML_DOC, 'div', {}, NotFound),
        (many.etree, HTML_DOC, 'div', {'allow_mismatch': True}, []),

        # anything with an 'xpath' method works
        (many.etree, MyTree(HTML_DOC), 'body/p/b/text()', {}, ['forban']),
        (many.etree, MyTree(HTML_DOC), 'p b', {}, ['<b>forban</b>!']),

        # pods matching
        (many.pods, {'number': 1926}, 'number', {}, [1926]),
        (many.pods, {'string': 's'}, 'string', {}, ['s']),
        (many.pods, {'dict': {'k': 'v'}}, 'dict', {}, [{'k': 'v'}]),
        (many.pods, {'null': None}, 'null', {}, [None]),
        (many.pods, {}, 'anything', {}, NotFound),
        (many.pods, {}, 'anything', {'allow_mismatch': True}, []),
        # if the needle resolves to a single list, we return the one list. Empty lists are fine
        (many.pods, {'list': []}, 'list', {}, [[]]),
        (many.pods, {'list': [1, 2, 3]}, 'list', {}, [[1, 2, 3]]),
        # if the needle ends in `[]`, however, meaning that we intent to select the list's elements, then the list must have
        # elements
        (many.pods, {'list': []}, 'list[]', {}, NotFound),
        (many.pods, {'list': []}, 'list[]', {'allow_mismatch': True}, []),
        (many.pods, {'list': [1]}, 'list[]', {}, [1]),
        (many.pods, {'list': [1, 2]}, 'list[]', {}, [1, 2]),
        (many.pods, {'list': [[1, 2]]}, 'list[]', {}, [[1, 2]]),
        (many.pods, [1, 2], '[]', {}, [1, 2]),
        (many.pods, [], '[]', {}, NotFound),
        (many.pods, {'list_of_obj': [{'v': 1}, {'v': 2}]}, 'list_of_obj', {}, [[{'v': 1}, {'v': 2}]]),
        (many.pods, {'list_of_obj': [{'v': 1}, {'v': 2}]}, 'list_of_obj[]', {}, [{'v': 1}, {'v': 2}]),
        (many.pods, {'list_of_obj': [{'v': 1}, {'v': 2}]}, 'list_of_obj[].v', {}, [1, 2]),
        (many.pods, {'list_of_obj': [{'v': 1}, {'other': 2}]}, 'list_of_obj[].v', {}, [1]),
        (many.pods, {'list_of_obj': [{'v': 1}, {'other': 2}]}, 'list_of_obj[].k', {}, NotFound),
        (many.pods, {'v': [[0,1], [2,3], [4,5]]}, 'v[][0]', {}, [0, 2, 4]),

        # arbitrary list filtering
        (many.filter, list(range(10)), lambda i: i % 3 == 0, {}, [0, 3, 6, 9]),
        (many.filter, list(range(10)), lambda i: i > 12, {}, NotFound),
        (many.filter, list(range(10)), lambda i: i > 12, {'allow_mismatch': True}, []),

        # list filtering takes no kwargs
        (many.filter, [1, 2, 3], lambda v: v, {'anything_at_all': True}, TypeError),
    ]
)
def test_find_all(find, haystack, needle, options, expected):
    try:
        results = find(needle, haystack, **options)
    except Exception as ex:  # anything at all, pylint: disable=broad-except
        if isinstance(expected, type) and issubclass(expected, Exception):
            assert isinstance(ex, expected)
        else:
            raise
    else:
        assert isinstance(results, list), results
        results = [
            ET.tostring(element, encoding=str).strip()
            if isinstance(element, ET._Element)  # pylint: disable=protected-access
            else element
            for element in results
        ]
        assert results == expected



#!/usr/bin/env python3

# standards
from typing import Any, Callable, Iterable, List, Optional, Tuple, Type, TypeVar, overload

# 3rd parties
from typing_extensions import Literal  # for pre-3.8 pythons

# poisk
from . import many
from .exceptions import ManyFound, NotFound
from .pods import SearchablePods
from .types import RegexType, XPathType


T = TypeVar('T')  # pylint: disable=invalid-name

TPrime = TypeVar('TPrime')


@overload
def re(
    needle: RegexType,
    haystack: str,
    parse: None = None,
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> str:
    """
    Find and return the only match of `needle` (a regex) in `haystack` (a str). If `parse` is None, and allow_mismatch is False,
    then we return a str.
    """

@overload
def re(
    needle: RegexType,
    haystack: str,
    parse: None = None,
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> Optional[str]:
    """
    When `parse` is None and `allow_mismatch` is True, then we return an `Optional[str]`.
    """

@overload
def re(
    needle: RegexType,
    haystack: str,
    parse: Callable[[str], T],
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> T:
    """
    When `parse` is given, then we return whatever type `parse` returns.
    """

@overload
def re(
    needle: RegexType,
    haystack: str,
    parse: Callable[[str], T],
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> Optional[T]:
    """
    When `parse` is not None, and `allow_mismatch` is True, then we return whatever type `parse` returns, or None.
    """

def re(needle, haystack, parse=None, *, allow_mismatch=False, allow_many=False, allow_duplicates=False, flags=0):
    return _one(
        needle,
        haystack,
        many.re(
            needle,
            haystack,
            parse,
            allow_mismatch=allow_mismatch,
            flags=flags,
        ),
        allow_many,
        allow_duplicates,
    )


@overload
def re_groups(
    needle: RegexType,
    haystack: str,
    parse: None = None,
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> Tuple[str, ...]:
    ...

@overload
def re_groups(
    needle: RegexType,
    haystack: str,
    parse: None = None,
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> Optional[Tuple[str, ...]]:
    ...

@overload
def re_groups(
    needle: RegexType,
    haystack: str,
    parse: Callable[[Tuple[str, ...]], T],
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> T:
    ...

@overload
def re_groups(
    needle: RegexType,
    haystack: str,
    parse: Callable[[Tuple[str, ...]], T],
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    flags: int = 0,
) -> Optional[T]:
    ...

def re_groups(needle, haystack, parse=None, *, allow_mismatch=False, allow_duplicates=False, allow_many=False, flags=0):
    return re(
        needle,
        haystack,
        parse,
        allow_mismatch=allow_mismatch,
        allow_many=allow_many,
        allow_duplicates=allow_duplicates,
        flags=flags,
    )

re_groups = re  # type: ignore


@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: None = None,
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    **kwargs
) -> XPathType:
    """
    `needle` is an XPath/CSS query. If `parse` is None, we return an Element. See notes at `many.etree`.
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: None = None,
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    **kwargs
) -> Optional[XPathType]:
    """
    When `parse=None` and `allow_mismatch=True`, we return an `Optional[Element]`
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: Callable[[XPathType], T],
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    **kwargs
) -> T:
    """
    When `parse` is not None, we return whatever type `parse` returns.
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: Callable[[XPathType], T],
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    **kwargs
) -> Optional[T]:
    """
    When `parse` is not None and `allow_mismatch=True`, we return whatever type `parse` returns, or None.
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: Callable[[str], T],
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
    **kwargs
) -> T:
    """
    When `parse` is a callable that accepts a `str`, we return whatever type `parse` returns. See notes at the parallel annotation
    for `many.etree`.
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: Callable[[str], T],
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
    **kwargs
) -> Optional[T]:
    """
    When `parse` is a callable that accepts a `str` and `allow_mismatch=True`, we return whatever type `parse` returns, or None.
    """

def etree(
    needle,
    haystack,
    parse=None,
    *,
    allow_mismatch=False,
    allow_many=False,
    allow_duplicates=False,
    **kwargs,
):
    return _one(
        needle,
        haystack,
        many.etree(
            needle,
            haystack,
            parse,
            allow_mismatch=allow_mismatch,
            **kwargs,
        ),
        allow_many,
        allow_duplicates,
    )


@overload
def attrib(
    needle: str,
    haystack: XPathType,
    parse: None = None,
    *,
    allow_mismatch: Literal[False] = False,
) -> str:
    ...

@overload
def attrib(
    needle: str,
    haystack: XPathType,
    parse: None = None,
    *,
    allow_mismatch: Literal[True],
) -> Optional[str]:
    ...

@overload
def attrib(
    needle: str,
    haystack: XPathType,
    parse: Callable[[str], T],
    *,
    allow_mismatch: Literal[False] = False,
) -> T:
    ...

@overload
def attrib(
    needle: str,
    haystack: XPathType,
    parse: Callable[[str], T],
    *,
    allow_mismatch: Literal[True],
) -> Optional[T]:
    ...

def attrib(needle, haystack, parse=None, *, allow_mismatch=False):
    try:
        value = haystack.attrib[needle]
    except KeyError as error:
        if allow_mismatch:
            return None
        else:
            raise NotFound(needle, haystack) from error
    if parse:
        return parse(value)
    return value


@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: None = None,
    *,
    type: None = None,
    allow_mismatch: Literal[False] = False,
) -> Any:
    """
    When `parse` is None and `type` is None, we return an object.
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: None = None,
    *,
    type: Type[T],
    allow_mismatch: Literal[False] = False,
) -> T:
    """
    If `type` is specified, we return that.
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: None = None,
    *,
    type: None = None,
    allow_mismatch: Literal[True],
) -> Optional[Any]:
    """
    If `allow_mismatch` is True and `type` is None, we return an `object`, or None
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: None = None,
    *,
    type: Type[T],
    allow_mismatch: Literal[True],
) -> Optional[T]:
    """
    If `type` is specified and `allow_mismatch=True`, we return a `type` instance, or None.
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: Callable[[Any], T],
    *,
    type: None = None,
    allow_mismatch: Literal[False] = False,
) -> T:
    """
    If `parse` is not None, we return whatever type `parse` returns.
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: Callable[[T], TPrime],
    *,
    type: Type[T],
    allow_mismatch: Literal[False] = False,
) -> TPrime:
    """
    If `parse` is set and `type` is set, then `parse` must accept an instance of `type`, and we return whatever `parse` returns
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: Callable[[Any], T],
    *,
    type: None = None,
    allow_mismatch: Literal[True],
) -> Optional[T]:
    """
    If `parse` is set and `allow_mismatch=True`, we return whatever type `parse` returns, or None.
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: Callable[[T], TPrime],
    *,
    type: Type[T],
    allow_mismatch: Literal[True],
) -> Optional[TPrime]:
    """
    If `type` is set, `parse` is set and `allow_mismatch=True`, `parse` must accept an instance of `type`, and we return whatever
    type `parse` returns, or None.
    """

def pods(
    needle,
    haystack,
    parse=None,
    *,
    type=None,
    allow_mismatch=False,
    allow_many=False,
    allow_duplicates=False,
):
    return _one(
        needle,
        haystack,
        many.pods(
            needle,
            haystack,
            parse,
            type=type,
            allow_mismatch=allow_mismatch,
        ),
        allow_many,
        allow_duplicates,
    )


@overload
def filter(
    needle: Callable[[T], object],
    haystack: Iterable[T],
    parse: None = None,
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
) -> T:
    """
    `haystack` is a sequence of T elements, and `needle` must accept `T` values. If `parse` is None, and allow_mismatch=False, we
    return a T.
    """

@overload
def filter(
    needle: Callable[[T], object],
    haystack: Iterable[T],
    parse: None = None,
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
) -> Optional[T]:
    """
    `haystack` is a sequence of T elements, and `needle` must accept `T` values. If `parse` is None, and allow_mismatch=True, we
    return a T, or None.
    """

@overload
def filter(
    needle: Callable[[T], object],
    haystack: Iterable[T],
    parse: Callable[[T], TPrime],
    *,
    allow_mismatch: Literal[False] = False,
    allow_many: bool = False,
    allow_duplicates: bool = False,
) -> TPrime:
    """
    If `parse` is not None, and `allow_mismatch=False`, then we return whatever type `parse` returns.
    """

@overload
def filter(
    needle: Callable[[T], object],
    haystack: Iterable[T],
    parse: Callable[[T], TPrime],
    *,
    allow_mismatch: Literal[True],
    allow_many: bool = False,
    allow_duplicates: bool = False,
) -> Optional[TPrime]:
    """
    If `parse` is not None, and `allow_mismatch=True`, then we return whatever type `parse` returns, or None.
    """

def filter(
    needle,
    haystack,
    parse=None,
    *,
    allow_mismatch=False,
    allow_many=False,
    allow_duplicates=False,
):
    return _one(
        needle,
        haystack,
        many.filter(
            needle,
            haystack,
            parse,
            allow_mismatch=allow_mismatch,
        ),
        allow_many,
        allow_duplicates,
    )


def _one(needle: object, haystack: object, results: List[T], allow_many: bool, allow_duplicates: bool):
    if not results:
        return None  # allow_mismatch must have been True
    if allow_duplicates:
        seen = set()
        deduplicated: List[T] = []
        for element in results:
            if element not in seen:
                seen.add(element)
                deduplicated.append(element)
        results = deduplicated
    if len(results) > 1 and not allow_many:
        raise ManyFound(needle, haystack)
    return results[0]



#!/usr/bin/env python3

# standards
import re as _re
from typing import Any, Callable, Iterable, List, Tuple, Type, TypeVar, overload

# 3rd parties
from cssselect import HTMLTranslator

# poisk
from .exceptions import NotFound
from .pods import SearchablePods, pods_search
from .types import RegexType, XPathType


_css_to_xpath = HTMLTranslator().css_to_xpath

_filter = filter


# type annotations used below

T = TypeVar('T')  # pylint: disable=invalid-name

TPrime = TypeVar('TPrime')


@overload
def re(
    needle: RegexType,
    haystack: str,
    parse: None = None,
    *,
    allow_mismatch: bool = False,
    flags: int = 0,
) -> List[str]:
    """
    Find and return all matches of `needle` in `haystack`. If `parse` is None, then we return a list of str's.
    """

@overload
def re(
    needle: RegexType,
    haystack: str,
    parse: Callable[[str], T],
    *,
    allow_mismatch: bool = False,
    flags: int = 0,
) -> List[T]:
    """
    When `parse` is not None, then we return a list of whatever type `parse` returns.
    """

def re(needle, haystack, parse=None, *, allow_mismatch=False, flags=0):
    results = _re.findall(needle, haystack, flags=flags)
    return _many(
        needle,
        haystack,
        results,
        parse,
        allow_mismatch=allow_mismatch,
    )

@overload
def re_groups(
    needle: RegexType,
    haystack: str,
    parse: None = None,
    *,
    allow_mismatch: bool = False,
    flags: int = 0,
) -> List[Tuple[str, ...]]:
    ...

@overload
def re_groups(
    needle: RegexType,
    haystack: str,
    parse: Callable[[Tuple[str, ...]], T],
    *,
    allow_mismatch: bool = False,
    flags: int = 0,
) -> List[T]:
    ...

def re_groups(needle, haystack, parse=None, *, allow_mismatch=False, flags=0):
    return re(needle, haystack, parse, allow_mismatch=allow_mismatch, flags=flags)

re_groups = re  # type: ignore


@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: None = None,
    *,
    allow_mismatch: bool = False,
    **kwargs
) -> List[XPathType]:
    """
    When `needle` is an XPath/CSS query. If `parse` is None, we return a list of Elements. Note that this means that if the xpath
    selects a string (e.g. "./a/@href"), then you need to specify `parse=str` to please the type checker.

    Note that the type annotation here is slightly off. It says we accept any type that has an `xpath` method, and then return a
    list of the same type, which isn't 100% correct: the output should be a list of ET._Element objects, but the input could be an
    ET._ElementTree object. Hopefully it'll work out, since the two classes have very similar interfaces.
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: Callable[[XPathType], T],
    *,
    allow_mismatch: bool = False,
    **kwargs
) -> List[T]:
    """
    When `parse` is not None, we return a list of whatever type `parse` returns. Use this with `parse=str` if the xpath selector
    selects text.
    """

@overload
def etree(
    needle: str,
    haystack: XPathType,
    parse: Callable[[str], T],
    *,
    allow_mismatch: bool = False,
    **kwargs
) -> List[T]:
    """
    When `parse` is a callable that accepts a `str`, we return a list of whatever type `parse` returns. Use this with e.g.
    `parse=int` to convert text selected by the xpath to an int. Note that it can still fail at runtime if the xpath selector
    doesn't return a string.
    """

def etree(needle, haystack, parse=None, *, allow_mismatch=False, **kwargs):
    if _re.search(r'[@/]|\(\)', needle):
        # XPath is able to search outside of a given node's subtree. We don't want that, we only want to change the subtree. If the
        # path doesn't already start with "./", prepend a dot, and slashes if there weren't already some.
        xpath = _re.sub(r'^(?!\./)/{,2}', lambda m: '.' + (m.group() or '//'), needle)
    else:
        xpath = _css_to_xpath(needle)
    results = haystack.xpath(xpath, **kwargs)
    return _many(
        needle,
        haystack,
        results,
        parse,
        allow_mismatch=allow_mismatch,
    )


@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: None = None,
    *,
    type: None = None,
    allow_mismatch: bool = False,
) -> List[Any]:
    """
    PODS search
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: None = None,
    *,
    type: Type[T],
    allow_mismatch: bool = False,
) -> List[T]:
    """
    If you add type=T, then we return a list of T
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: Callable[[Any], T],
    *,
    type: None = None,
    allow_mismatch: bool = False,
) -> List[T]:
    """
    If `parse` is not None, we return a list of whatever type `parse` returns.
    """

@overload
def pods(
    needle: str,
    haystack: SearchablePods,
    parse: Callable[[T], TPrime],
    *,
    type: Type[T],
    allow_mismatch: bool = False,
) -> List[TPrime]:
    """
    You can again check for a specific type using the `type` kwargs. If `parse` is set, then it must accept an instance of `type`.
    """

def pods(needle, haystack, parse=None, *, type=None, allow_mismatch=False):
    results = pods_search(needle, haystack, type=type)
    return _many(
        needle,
        haystack,
        results,
        parse,
        allow_mismatch=allow_mismatch,
    )


@overload
def filter(
    needle: Callable[[T], object],
    haystack: Iterable[T],
    parse: None = None,
    *,
    allow_mismatch: bool = False,
) -> List[T]:
    """
    `haystack` is a sequence of T elements, and `needle` must accept `T` values. If `parse` is None, we return T's.
    """

@overload
def filter(
    needle: Callable[[T], object],
    haystack: Iterable[T],
    parse: Callable[[T], TPrime],
    *,
    allow_mismatch: bool = False,
) -> List[TPrime]:
    """
    If `needle` is callable, and `parse` is not None, then we return a list of whatever type `parse` returns.
    """

def filter(needle, haystack, parse=None, allow_mismatch=False):
    results = list(_filter(needle, haystack))
    return _many(
        needle,
        haystack,
        results,
        parse,
        allow_mismatch=allow_mismatch,
    )


def _many(needle, haystack, results, parse=None, allow_mismatch=False):
    if not results and not allow_mismatch:
        raise NotFound(needle, haystack)
    if parse is not None:
        results = list(map(parse, results))
    return results



#!/usr/bin/env python3


class PoiskException(ValueError):

    def __init__(self, needle, haystack):
        super().__init__(self._compose_message(needle, haystack))
        self.needle = needle
        self.haystack = haystack

    @staticmethod
    def _compose_message(needle, haystack):
        message = repr(needle)
        if isinstance(haystack, (str, list, tuple, set, dict)):
            haystack_repr = repr(haystack)
            if len(haystack_repr) > 100:
                haystack_repr = haystack_repr[:50] + '…' + haystack_repr[50:-1]
            message += ' in ' + haystack_repr
        return message


class NotFound(PoiskException):
    pass


class ManyFound(PoiskException):
    pass



#!/usr/bin/env python3

# standards
from pathlib import Path
import re
import setuptools


README_FILE = Path(__file__).parent / 'README.md'

LONG_DESCRIPTION = README_FILE.read_text('UTF-8')

LONG_DESCRIPTION = re.sub(
    r'(?<=\]\()(?!http)',
    'https://github.com/saintamh/poisk/tree/master/',
    LONG_DESCRIPTION,
)


setuptools.setup(
    name='poisk',
    version='1.0.1',
    description='Small utilities for searching data structures',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Hervé Saint-Amand',
    author_email='poisk@saintamh.org',
    package_data={'poisk': ['py.typed']},
    packages=setuptools.find_packages(),
    install_requires=[
        'cssselect>=1.1,<2',
        'typing-extensions>=3.7,<4',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    zip_safe=False,  # https://mypy.readthedocs.io/en/latest/installed_packages.html#creating-pep-561-compatible-packages
)



#!/usr/bin/env python3

# standards
import re
from typing import Any, Callable, TypeVar, Union

# 3rd parties
from typing_extensions import Protocol  # for pre-3.8 pythons


class HasXPathMethod(Protocol):

    @property
    def xpath(self) -> Callable[..., Any]:
        # This declaration here is not actually for a property, it's for a method called `xpath`, whose signature we don't care
        # about (beyond its name). Apparently this is the way you can get mypy to recognize this. This hacky solution of using a
        # @property was coined (by the BDFL) here: https://github.com/python/mypy/issues/9560#issuecomment-705955103
        ...


XPathType = TypeVar('XPathType', bound=HasXPathMethod)


RegexType = Union[str, re.Pattern]
